#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # Mapper, Reducers customisés avec SQL
# 
# Ce notebook propose l'utilisation de SQL avec [SQLite](https://sqlite.org/) pour manipuler les données depuis un notebook (avec le module [sqlite3](https://docs.python.org/3.6/library/sqlite3.html)).

# In[1]:


from jyquickhelper import add_notebook_menu
add_notebook_menu()


# In[2]:


get_ipython().run_line_magic('matplotlib', 'inline')
import matplotlib.pyplot as plt
plt.style.use('ggplot')
import pyensae
from pyquickhelper.helpgen import NbImage


# ## Représentation

# Le module [pandas](http://pandas.pydata.org/) manipule des tables et c'est la façon la plus commune de représenter les données. Lorsque les données sont multidimensionnelles, on distingue les coordonnées des valeurs :

# In[3]:


NbImage("cube1.png")


# Dans cet exemple, il y a :
#     
# * 3 coordonnées : Age, Profession, Annéee
# * 2 valeurs : Espérance de vie, Population
# 
# On peut représenter les donnés également comme ceci :

# In[4]:


NbImage("cube2.png")


# C'est assez simple. Prenons un exemple : [table de mortalité de 1960 à 2010](http://www.data-publica.com/opendata/7098--population-et-conditions-sociales-table-de-mortalite-de-1960-a-2010) qu'on récupère à l'aide de la fonction [table_mortalite_euro_stat](http://www.xavierdupre.fr/app/actuariat_python/helpsphinx/actuariat_python/data/population.html#actuariat_python.data.population.table_mortalite_euro_stat). C'est assez long (4-5 minutes) sur l'ensemble des données car elles doivent être prétraitées (voir la documentation de la fonction). Pour écouter, il faut utiliser le paramètre *stop_at*.

# In[5]:


from sparkouille.datasets import table_mortalite_euro_stat 
table_mortalite_euro_stat()


# In[6]:


import os
os.stat("mortalite.txt")


# In[7]:


import pandas
df = pandas.read_csv("mortalite.txt", sep="\t", encoding="utf8", low_memory=False)
df.head()


# Les indicateurs pour deux âges différents :

# In[8]:


df[ ((df.age=="Y60") | (df.age=="Y61")) & (df.annee == 2000) & (df.pays=="FR") & (df.genre=="F")]


# ## Données trop grosses pour tenir en mémoire : SQLite
# 
# On charge une grosse base de données (assez petite pour que la séance ne soit pas trop longue).

# In[9]:


df.shape


# Les données sont trop grosses pour tenir dans une feuille Excel et les consulter il n'y a pas d'autres moyens que d'en regarder des extraits. Que passe-t-il quand les données sont encore plus grosses et qu'elles ne tiennent pas en mémoire ? Quelques solutions :
# 
# * augmenter la mémoire de l'ordinateur, avec 20 Go, on peut faire beaucoup de choses,
# * stocker les données dans un serveur SQL,
# * stocker les données sur un système distribué (cloud, Hadoop, ...)
# 
# La seconde option n'est pas toujours simple, il faut installer un serveur SQL. Pour aller plus vite, on peut simplement utiliser [SQLite](https://www.sqlite.org/) qui est une façon de faire du SQL sans serveur (cela prend quelques minutes). On utilise la méthode [to_sql](http://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.to_sql.html).

# In[10]:


import sqlite3
from pandas.io import sql
cnx = sqlite3.connect('mortalite.db3')
try:
    df.to_sql(name='mortalite', con=cnx)
except ValueError as e:
    if "Table 'mortalite' already exists" not in str(e):
        # seulement si l'erreur ne vient pas du fait que cela 
        # a déjà été fait
        raise e
# on peut ajouter d'autres dataframe à la table comme si elle était créée par morceau
# voir le paramètre if_exists de la fonction to_sql


# On peut maintenant récupérer un morceau avec la fonction [read_sql](http://pandas.pydata.org/pandas-docs/stable/generated/pandas.read_sql.html?highlight=read_sql#pandas.read_sql).

# In[11]:


import pandas
example = pandas.read_sql('SELECT * FROM mortalite WHERE age_num==50 LIMIT 5', cnx)
example


# L'ensemble des données restent sur le disque, seul le résultat de la requête est chargé en mémoire. Si on ne peut pas faire tenir les données en mémoire, il faut soit en obtenir une vue partielle (un échantillon aléatoire, un vue filtrée), soit une vue agrégrée. Pour finir, il faut fermer la connexion pour laisser d'autres applications ou notebook modifier la base ou tout simplement supprimer le fichier.

# In[12]:


cnx.close()


# Sous Windows, on peut consulter la base avec le logiciel [SQLiteSpy](http://www.yunqa.de/delphi/doku.php/products/sqlitespy/index).

# In[13]:


NbImage("sqlite.png")


# Sous Linux ou Max, on peut utiliser une extension Firefox [SQLite Manager](https://addons.mozilla.org/en-US/firefox/addon/sqlite-manager/?src). Dans ce notebook, on utilisera la commande magique [%%SQL](http://www.xavierdupre.fr/app/pyensae/helpsphinx/notebooks/pyensae_sql_magic.html) du module [pyensae](http://www.xavierdupre.fr/app/pyensae/helpsphinx/) : 

# In[14]:


get_ipython().run_line_magic('load_ext', 'pyensae')
get_ipython().run_line_magic('SQL_connect', 'mortalite.db3')


# In[15]:


get_ipython().run_line_magic('SQL_tables', '')


# In[16]:


get_ipython().run_line_magic('SQL_schema', 'mortalite')


# In[17]:


get_ipython().run_cell_magic('SQL', '', 'SELECT COUNT(*) FROM mortalite\n')


# In[18]:


get_ipython().run_line_magic('SQL_close', '')


# ## Cas 1 : filtrer pour créer un échantillon aléatoire
# 
# Si on ne peut pas faire tenir les données en mémoire, on peut soit regarder les premières lignes soit prendre un échantillon aléatoire. Deux options :
# 
# * [Dataframe.sample](http://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.sample.html)
# * [create_function](https://docs.python.org/3.4/library/sqlite3.html#sqlite3.Connection.create_function)
# 
# La première fonction est simple :

# In[19]:


sample = df.sample(frac=0.1)
sample.shape, df.shape


# Je ne sais pas si cela peut être réalisé sans charger les données en mémoire. Si les données pèsent 20 Go, cette méthode n'aboutira pas. Pourtant, on veut juste un échantillon pour commencer à regarder les données. On utilise la seconde option avec [create_function](https://docs.python.org/3.4/library/sqlite3.html#sqlite3.Connection.create_function) et la fonction suivante :

# In[20]:


import random  #loi uniforme
def echantillon(proportion):
    return 1 if random.random() < proportion else 0


# In[21]:


import sqlite3
from pandas.io import sql
cnx = sqlite3.connect('mortalite.db3')


# On déclare la fonction à la base de données.

# In[22]:


cnx.create_function('echantillon', 1, echantillon)


# On veut récupérer environ 1% de la table ? On écrit d'abord le filtre.

# In[23]:


sample = pandas.read_sql('SELECT * FROM mortalite WHERE echantillon(0.01)', cnx)
sample.shape


# In[24]:


sample.head()


# On ferme la connexion.

# In[25]:


cnx.close()


# ## Pseudo Map/Reduce avec SQLite

# La liste des [mots-clés du langage SQL utilisés par SQLite](https://www.sqlite.org/keyword_index.html) n'est pas aussi riche que d'autres solutions de serveurs SQL. La médiane ne semble pas en faire partie. Cependant, pour une année, un genre, un âge donné, on voudrait calculer la médiane de l'espérance de vie sur l'ensembles des pays.

# In[26]:


import sqlite3, pandas
from pandas.io import sql
cnx = sqlite3.connect('mortalite.db3')


# In[27]:


pays = pandas.read_sql('SELECT pays, COUNT(*) FROM mortalite GROUP BY pays', cnx)
pays.head()


# Il n'y a pas le même nombre de données selon les pays, il est probable que le nombre de pays pour lesquels il existe des données varie selon les âges et les années.

# In[28]:


query = """SELECT nb_country, COUNT(*) AS nb_rows FROM (
                SELECT annee,age,age_num, count(*) AS nb_country FROM mortalite 
                WHERE indicateur=="LIFEXP" AND genre=="F"
                GROUP BY annee,age,age_num
            ) GROUP BY nb_country"""
df = pandas.read_sql(query, cnx)


# In[29]:


df.sort_values("nb_country", ascending=False).head(n=2)


# In[30]:


ax = df.plot(x="nb_country", y="nb_rows")
ax.set_title("Nombre de données par pays");


# Soit un nombre inconstant de pays. Le fait qu'on est 100 pays suggère qu'on ait une erreur également.

# In[31]:


query = """SELECT annee,age,age_num, count(*) AS nb_country FROM mortalite 
                WHERE indicateur=="LIFEXP" AND genre=="F"
                GROUP BY annee,age,age_num
                HAVING nb_country >= 100"""
df = pandas.read_sql(query, cnx)


# In[32]:


df.head()


# Ce sont des valeurs manquantes. Le problème pour calculer la médiane pour chaque observation est qu'il faut d'abord regrouper les lignes de la table par indicateur puis choisir la médiane dans chaque de ces petits groupes. On s'inspire pour cela de la logique Map/Reduce et de la fonction [create_aggregate](https://docs.python.org/3.4/library/sqlite3.html#sqlite3.Connection.create_aggregate).

# ## Cas 2 : reducer customisé avec SQL
# 
# Le reducer se présente toujours sous la forme suivante :

# In[33]:


class ReducerMediane:
    def __init__(self):
        # ???
        pass
    def step(self, value):
        # ???
        #
        pass
    def finalize(self):
        # ???
        # return ... //2 ]
        pass


# Qu'on renseigne de la sorte :

# In[34]:


class ReducerMediane:
    def __init__(self):
        self.indicateur = []
    def step(self, value):
        if value >= 0:
            self.indicateur.append(value)
    def finalize(self):
        self.indicateur.sort()
        return self.indicateur[len(self.indicateur)//2]


# On le déclare ensuite à *sqllite3*.

# In[35]:


cnx.create_aggregate("ReducerMediane", 1, ReducerMediane) 


# In[36]:


query = """SELECT annee,age,age_num, ReducerMediane(valeur) AS mediane FROM mortalite 
                WHERE indicateur=="LIFEXP" AND genre=="F"
                GROUP BY annee,age,age_num"""
df = pandas.read_sql(query, cnx)


# In[37]:


df.head()


# Un reducer à deux entrées même si cela n'a pas beaucoup de sens ici :

# In[38]:


class ReducerMediane2:
    def __init__(self):
        self.indicateur = []
    def step(self, value, value2):
        if value >= 0:
            self.indicateur.append(value)
        if value2 >= 0:
            self.indicateur.append(value2)
    def finalize(self):
        self.indicateur.sort()
        return self.indicateur[len(self.indicateur)//2]
    
cnx.create_aggregate("ReducerMediane2", 2, ReducerMediane2)


# In[39]:


query = """SELECT annee,age,age_num, ReducerMediane2(valeur, valeur+1) AS mediane2 FROM mortalite 
                WHERE indicateur=="LIFEXP" AND genre=="F"
                GROUP BY annee,age,age_num"""
df = pandas.read_sql(query, cnx)
df.head()


# Il n'est apparemment pas possible de retourner deux résultats mais on peut utiliser une ruse qui consise à les concaténer dans une chaîne de caracères.

# In[40]:


class ReducerQuantile:
    def __init__(self):
        self.indicateur = []
    def step(self, value):
        if value >= 0:
            self.indicateur.append(value)
    def finalize(self):
        self.indicateur.sort()
        q1 = self.indicateur[len(self.indicateur)//4]
        q2 = self.indicateur[3*len(self.indicateur)//4]
        n = len(self.indicateur)
        return "%f;%f;%s" % (q1,q2,n)
    
cnx.create_aggregate("ReducerQuantile", 1, ReducerQuantile)


# In[41]:


query = """SELECT annee,age,age_num, ReducerQuantile(valeur) AS quantiles FROM mortalite 
                WHERE indicateur=="LIFEXP" AND genre=="F"
                GROUP BY annee,age,age_num"""
df = pandas.read_sql(query, cnx)
df.head()


# On ferme la connexion.

# In[42]:


cnx.close()


# ## Notion d'index

# En SQL et pour de grandes tables, la notion d'index joue un rôle important pour accélérer les opérations de jointures (``JOIN``) ou de regroupement (``GROUP BY``). L'article [A thorough guide to SQLite database operations in Python](http://sebastianraschka.com/Articles/2014_sqlite_in_python_tutorial.html) montre comment faire les principales opérations.

# In[43]: